
#include "module.h"

// Module implementation
float Module::samplerate = 44100;
long long Module::time = 0;

float Module::Out() {
    if (time > last_time) {
        Update();
        last_time = time;
    }
    return output;
}


// Noise implementation
void Noise::Update() {
    output = 2.0 * (float)rand() / RAND_MAX - 1;
}


// Mixer2 implementation
void Mixer2::Update() {
    output = gain1 * input1->Out() + gain2 * input2->Out();
}

// Amplifier implementation
void Amplifier::Update() {
    output = input->Out() * gain * (modulator ? modulator->Out() : 1.0f);
}


// SineOsc implementation
void SineOsc::Update() {
    float freq = frequency * (1 + (frequency_modulator ? (frequency_modulator->Out() * modulation_amount) : 0));
    float dt = (float)(time - last_time) / samplerate;
    phase += dt * freq;
    output = cos(phase * 2 * M_PI);
}


// LinearSlew implementation
void LinearSlew::SetTargetTime(float level, float delay) {
    target = level;
    delta = (target - output) / (delay * samplerate);
    target_reached = false;
}

void LinearSlew::SetTargetRate(float level, float rate) {
    target = level;
    delta = rate / samplerate;
    target_reached = false;
}

void LinearSlew::Update() {
    output += delta;
    target_reached = (((delta > 0) - (delta < 0)) * (output - target)) >= 0;
    delta = delta * !target_reached;
//    switch (sgn(delta)) {
//        case -1: if (output - target <= 0) delta = 0; break;
//        case 1: if (output - target >= 0) delta = 0; break;
//    }
}


// AREnvelope implementation
void AREnvelope::Update() {
    LinearSlew::Update();
    if (gate) {
        if (!release_stage) {
            if (target_reached) {
                release_stage = true;
                SetTargetTime(0, release);
            }
        }
        else {
            if (target_reached) {
                gate = release_stage = false;
            }
        }
    }
    else {
        if ((trigger_time > 0) && (time >= trigger_time)) {
            trigger_time = -1;
            if (retrigger) output = 0;
            gate = true;
            release_stage = false;
            SetTargetTime(1, attack);
        }
    }
}



// Delay implementation
Delay::Delay(float max_time) : buffer(NULL), buffer_size(0), write_pos(0), input(NULL), delay_time(1), feedback(0), mix(1.0f) {
    buffer_size = (int)(fabs(max_time) * samplerate + 0.5);
    buffer = new float[buffer_size];
}

Delay::~Delay() {
    if (buffer) delete[] buffer;
    buffer = NULL;
}

void Delay::SetTime(float t) {
    delay_time = (int)(t * samplerate);
}

void Delay::Update() {
    float in, out;
    int read_pos = write_pos - delay_time;
    in = input->Out();
    read_pos += buffer_size * (read_pos < 0);
    out = buffer[read_pos];
    buffer[write_pos++] = in + feedback * out;
    write_pos -= buffer_size * (write_pos >= buffer_size);
    output = (1.0f - mix) * in + mix * out;
}


