
#include <windows.h>
#include <iomanip>
#include <cctype>
#include "synth.h"

#include <conio.h>

// GUIControl Implementation

GUIControl::GUIControl() :
    parameter(NULL),
    min(0),
    max(0),
    inckey(0),
    deckey(0),
    size(0),
    delta(0) {
    strcpy(name, "");
}

GUIControl::GUIControl(const char *_n, float *_p, float _min, float _max, char _ik, char _dk, int s) :
    parameter(_p),
    min(_min),
    max(_max),
    inckey(_ik),
    deckey(_dk),
    size(s) {
    strncpy(name, _n, 12);
    name[12] = 0;
    if (min > max) { delta = max; max = min; min = delta; }
    delta = (max - min) / size;
}

float GUIControl::GetDelta() {
    return delta;
}

float GUIControl::GetValue() {
    return parameter ? *parameter : 0;
}

void GUIControl::SetValue(float v) {
    if (parameter && (v >= min) && (v <= max)) *parameter = v;
}

void GUIControl::AddValue(float v) {
    SetValue(*parameter + v);
}

ostream &operator <<(ostream &os, GUIControl &g) {
    int i, n;
    n = (int)((g.GetValue() - g.min) / g.delta + 0.5);
    os << setw(12) << g.name << " - " << g.deckey << "[";
    for (i = 0; i < g.size; i++) {
        os << ((i < n) ? '*' : ' ');
    }
    os << "]" << g.inckey;
    return os;
}

// KeyCommand Implementation

KeyCommand::KeyCommand() : control(NULL), delta(0) {
}

void KeyCommand::Set(GUIControl *_c, float d) {
    control = _c;
    delta = d;
}

void KeyCommand::Execute() {
    if (control) control->AddValue(delta);
}





// Synth implementation

static int paCallback(const void *inputBuffer, void *outputBuffer, unsigned long framesPerBuffer,
                        const PaStreamCallbackTimeInfo *timeInfo, PaStreamCallbackFlags statusFlags, void *userData) {
    unsigned int i;
    Module *module = (Module *)userData;
    float *out = (float *)outputBuffer;
    float sample;
    if (module) {
        for (i = 0; i < framesPerBuffer; i++) {
            sample = module->Out();
            Module::IncTime();
            *out++ = sample;
            *out++ = sample;
        }
    } else {
        memset(out, 0, framesPerBuffer * 2 * sizeof(float));
    }
    return 0;
}


Synth::Synth() : delay(10) {
    PaError err;

    // connect modules
    delay.input = &carrier_amp;
    carrier_amp.input = &carrier;
    carrier_amp.modulator = &carrier_env;
    carrier.frequency_modulator = &modulator_amp;
    modulator_amp.input = &modulator;
    modulator_amp.modulator = &modulator_env;
    modulator.frequency_modulator = &modulator_lfo;

    // Initial (test) parameters
    carrier_env.attack = 0.005;
    carrier_env.release = 0.5;
    modulator_env.attack = 0.1;
    modulator_env.release = 0.5;
    carrier.frequency = 220;
    modulator.frequency = 440;
    carrier.modulation_amount = 1;
    modulator_lfo.frequency = 0.1;
    modulator.modulation_amount = 0.3;
    delay.SetTime(0.5);
    delay.feedback = 0.2;
    delay.mix = 0.2;

    // Build interface
    numcontrols = 0;
    AddControl("FM Depth", &carrier.modulation_amount, 0, 4, 'Q', 'A');
    AddControl("ModEnv Attack", &modulator_env.attack, 0, 10, 'W', 'S');
    AddControl("ModEnv Release", &modulator_env.release, 0, 10, 'E', 'D');
    AddControl("CarEnv Attack", &carrier_env.attack, 0, 10, 'R', 'F');
    AddControl("CarEnv Release", &carrier_env.release, 0, 10, 'T', 'G');
    AddControl("LFO Freq", &modulator_lfo.frequency, 0, 10, 'Y', 'H');
    AddControl("LFO Depth", &modulator.modulation_amount, 0, 2, 'U', 'J');
    AddControl("Delay Feedback", &delay.feedback, 0, 1, 'I', 'K');
    AddControl("Delay Mix", &delay.mix, 0, 1, 'O', 'L');

    // Initialize audio
    err = Pa_Initialize();
    if (err != paNoError) exit(-1);

    // Initialize audio stream
    err = Pa_OpenDefaultStream(&stream, 0, 2, paFloat32, Module::GetSampleRate(), 256, paCallback, &delay);
    if (err != paNoError) exit(-1);

    // Start playback
    err = Pa_StartStream(stream);
    if (err != paNoError) exit(-1);
    Pa_Sleep(100);
}

Synth::~Synth() {
    PaError err;

    // Terminate audio stream
    err = Pa_StopStream(stream);
    if (err != paNoError) exit(-1);

    err = Pa_CloseStream(stream);
    if (err != paNoError) exit(-1);

    err = Pa_Terminate();
    if (err != paNoError) exit(-1);
}


void Synth::AddControl(const char *_n, float *_p, float _min, float _max, char _ik, char _dk, int s) {
    if (numcontrols < 20) {
        control[numcontrols] = GUIControl(_n, _p, _min, _max, _ik, _dk, s);
        keycommand[(int)_ik].Set(&control[numcontrols], control[numcontrols].GetDelta());
        keycommand[(int)_dk].Set(&control[numcontrols], -control[numcontrols].GetDelta());
        numcontrols++;
    }
}

void Synth::ShowGUI() {
    int i;
    system("cls"); // system("clear") on linux/osx
    for (i = 0; i < numcontrols; i++) {
        cout << control[i] << endl;
    }
}



void Synth::PlayNote(int midinote) {
    double freq = 440.0 * pow(2.0, (double)(midinote - 57) / 12.0);
    carrier.frequency = freq;
    modulator.frequency = freq * 2;
    carrier_env.trigger_time = Module::GetTime() + 1;
    modulator_env.trigger_time = Module::GetTime() + 1;
}


void Synth::Run() {
    int i;
    char c;

    const char notes[12] = "<zxcvbnm,.-";
    char midinotes[11] = { 0, 2, 4, 5, 7, 9, 11, 12, 14, 16, 17 };

    bool run = true;
    while (run) {
        if (kbhit()) {
            c = getch();
            if (c < 128) {
                keycommand[(int)toupper(c)].Execute();
                ShowGUI();
            }

            for (i = 0; i < 12; i++) {
                if (c == notes[i]) {
                    PlayNote(midinotes[i] + 60);
                    break;
                }
            }

            if (c == 27) run = false;
        }
        Sleep(0);
    }
}
